-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}

M._serialize = {
    differentialMode = true,
    manualShift = true,
    autoClutch = true,
    shifterMode = true,
    activeModeIndex = true
}

local warningState = {}

M.torque = 0
M.gear = 0
M.rpm = 0
M.torqueTransmission = 0
M.brake = 0
M.throttle = 0
M.throttleFactor = 1
M.wheels = {}
M.wheelInfo = {}
M.engineDisabled = false
M.engineIgnition = true
M.absPulse = 0
M.fuelLeakRate = 0 -- L/sec
M.fuel = 100
M.fuelCapacity = 100
M.engineLoad = 0

-- shifterMode = 0 : full manual
-- shifterMode = 1 : manual with automatic clutch
-- shifterMode = 2 : simple automatic
-- shifterMode = 3 : real automatic
M.shifterMode = 2 -- initial value needs to fit to the two modes below
M.differentialMode = 0
M.autoClutch = true
M.manualShift = false
M.shifterPosition = 0
M.wheelCount = 0
M.avgAV = 0
M.axleAngle = 0
M.driveshaftAngle = 0
M.clutch = 1
M.activeModeIndex = 1
M.engineAV = 0
M.engineThrottle = 0
M.engineWorkPerUpdate = 0
M.frictionLossPerUpdate = 0
M.pumpingLossPerUpdate = 0
M.driveshaftDisabled = false
M.brakingCoef = 0
M.engFriction = 20

local announcedShifterMode = nil
local diffs = {}
local diffCount = 1
local diffTypeMap = {open = 0, lsd = 1}
local diffStateMap = {open = 0, closed = 1, locked = 2}
local avgDiffAV = 0
local invDiffCount = 1
local invWheelCount = 0
local propWheelsCount = 0
local engInertia = 0.1
local invEngInertia = 10
local halfInvEngInertia = invEngInertia * 0.5
local axleAV = 0
local burnEfficiency = 1
local axleFriction = 50
local clutchTimer = 0
local clutchDuration = 0.5
local clutchRatio = 1
local clutchTorque = 0
local invClutchDuration = 2
local viscousCoupling
local viscousCouplingOrig
local engIdleAV = 0
local engStallAV = 0
local torqueReactionNodes = nil
local invBurnEfficiencyFuelDensity = 0
local torsionCoef = 0
local diffRatio = 2
local invDiffRatio = 0.5
local avgPropAV = 0
local maxGearRatio = 0
local gearChangedTimer = 0
local gearChangeDelay = 0.5
local idleTorque = 0
local maxTorque = 0
local maxPower = 0
local checkEngineTimer = 0
local engineDifferential = 1
local allModes = {0,1,2,3,4}
local autoModes = {2,3,4}
local manualModes = {2,1,0}
local activeModes = allModes
local avgAVsmoother = newExponentialSmoothing(15)
local loadSmoother = newTemporalSmoothing(1,1)
local smoothedAvgAV = 0
local floodLevel = 0 -- zero to one
local prevFloodPercent = 0
local hydrolockThreshold = 0.9
local freeze = 0
local ABSavgAV = 0
local engineState = 1
local revLimiterActive = false
local revLimiterAVThreshold = 0
local fuelNodes = {}
local fuelNodeWeightCoef = 0
local wheelradius = 0
local driveshaftBeamBreakLookup = {}
local engineBeamBreakLookup = {}
local fuelTankBeamBreakLookup = {}

local function nop()
end

local function sendWarningStateToUi ()
    guihooks.trigger('WarningState', {module = 'drivetrain', data = warningState})
end

M.engine = {updateGFX = nop, init = nop, beambroke = nop}
M.wheelSimulation = {updateGFX = nop, init = nop}
M.esc = {update = nop, updateGFX = nop, toggleESCMode = nop, getCarData = nop, escPulse = 0}
M.turbocharger = {updateGFX = nop, turboCoefPower = 1}

local function scaleEngineFriction(friction)
    M.engFriction = M.engFriction * friction
end

local function scaleEngineState(state)
    engineState = engineState * state
end

local function updateEngineState(dt)
    local max = math.max
	local blockTorque
	local lastAV
    local engine = v.data.engine
    local torquediff
    local torque = nil
	local lockup = engine.enableLocking

    --Calc wheel sided engine AV
    local wheelBasedEngAV = avgDiffAV * engine.gears[M.gear]

    local internal_clutchRatio = clutchRatio
    -- manual vs automatic clutch
    if M.autoClutch then
        -- auto clutch
        if clutchTimer > 0 then
            clutchTimer = max(0, clutchTimer - dt)
            internal_clutchRatio = math.min((clutchDuration - clutchTimer) * invClutchDuration, clutchRatio)
        end
    end

    --Calc torque from engine rpm
    local rpm = M.engineAV * 9.5493
    torque = (engine.torqueCurve[math.floor(rpm)] or 0)
    torque = torque * math.min(M.throttle * maxPower / (torque * rpm + 1e-30), 1) * M.throttleFactor * M.turbocharger.turboCoefPower

    --Idle torque
    if M.engineAV < engIdleAV and M.fuel ~= 0 then
        torque = max(torque, idleTorque)
    end

    torque = torque * engineState

    -- Calculate engine's work
    local dtT = dt * torque
    M.engineWorkPerUpdate = M.engineWorkPerUpdate + dtT * (dtT * halfInvEngInertia + M.engineAV)
    M.frictionLossPerUpdate = M.frictionLossPerUpdate + M.engFriction * M.engineAV * dt
    M.pumpingLossPerUpdate = M.pumpingLossPerUpdate + M.brakingCoef * M.engineAV * M.engineAV * dt

    -- decouple when in neutral
    if M.gear == 0 then
        torquediff = 0
        M.torque = 0
        M.torqueTransmission = 0
		lastAV = M.engineAV
        M.engineAV = max(M.engineAV + dt * (torque - M.engFriction - M.brakingCoef * M.engineAV) * invEngInertia, 0)
        axleAV = avgDiffAV
		blockTorque = (M.engineAV - lastAV) / dt * 0.5
		if torqueReactionNodes ~= nil then
				obj:applyTorque(torqueReactionNodes[1], torqueReactionNodes[2], torqueReactionNodes[3], blockTorque)
		end
    else
		-- torque converter implementation	
		if M.shifterMode == 4 then
			-- create converter specs from viscous coupling, screw it, lets put some safeties in here, too
			local converterStallAV
			if viscousCoupling < 6 then		--cause reasons
				converterStallAV = 6.0 / viscousCoupling * 209.4394 -- based on average 2000rpm stall, higher viscous value=lower stall
			else
				converterStallAV = 13.0 / viscousCoupling * 209.4394
			end
			-- convertor stall min 1800 rpm max 2800 rpm
			if converterStallAV > 293 then
				converterStallAV = 293
			elseif converterStallAV < 188 then
				converterStallAV = 188
			end
			local multiRatio = converterStallAV / 100.0 -- based on avg ~2x multiplier for car transmissions
			-- even more safety
			if multiRatio > 2.5 then
				multiRatio = 2.5
			elseif multiRatio < 1.7 then
				multiRatio = 1.7
			end
			--rpm = AV * 9.5493
			local speedratio
			local torqueratio
			-- getting speed ratio
			if wheelBasedEngAV < 0.0 then
			speedratio = 0.0
			else
			speedratio = wheelBasedEngAV/M.engineAV
			end
			
			-- modeling torque ratio
			-- if engine idling, limit torque
			if speedratio < 0.9 then
				torqueratio = multiRatio - ((multiRatio/1.4/0.9) * speedratio)
			elseif speedratio <= 1 then
				torqueratio = 1 * speedratio
			else
				torqueratio = -1 * speedratio
			end
			--check torque ratio
			if torqueratio > multiRatio then
				torqueratio = multiRatio
			elseif torqueratio < 1 then
				torqueratio = 1
			end
			-- stall speed?
			internal_clutchRatio = M.engineAV / converterStallAV
			-- Safety third...
			if internal_clutchRatio < 0.0 then
				internal_clutchRatio = 0
			elseif internal_clutchRatio > 1 then
				internal_clutchRatio = 1
			end
			-- lockup?
			
			if M.throttle <= 0.7 and M.gear >= 3 and lockup then
				torquediff = (M.engineAV - wheelBasedEngAV) * math.min(clutchTorque / maxGearRatio, (engInertia / obj:getPhysicsDt()))
			else
				torquediff = (M.engineAV - wheelBasedEngAV) * viscousCoupling
			end
			if math.abs(torquediff) > clutchTorque then
				torquediff = fsign(torquediff) * clutchTorque
			end
			torquediff = torquediff * internal_clutchRatio
			M.engineAV = max(M.engineAV + dt * (torque - torquediff - M.engFriction - M.brakingCoef * M.engineAV) * invEngInertia, 0)
			M.torque = torquediff
			M.torqueTransmission = torquediff * torqueratio * engine.gears[M.gear]
			--apply reaction torque to engine if under power
			if torqueReactionNodes ~= nil then
				if torquediff > 0 then
					obj:applyTorque(torqueReactionNodes[1], torqueReactionNodes[2], torqueReactionNodes[3], M.torqueTransmission * 0.4)
				end
			end
			axleAV = M.engineAV / engine.gears[M.gear]
		else
			--classic transmission
			torquediff = (M.engineAV - wheelBasedEngAV) * viscousCoupling
			if math.abs(torquediff) > clutchTorque then
				torquediff = fsign(torquediff) * clutchTorque
			end
			torquediff = torquediff * internal_clutchRatio
			M.engineAV = max(M.engineAV + dt * (torque - torquediff - M.engFriction - M.brakingCoef * M.engineAV) * invEngInertia, 0)

			M.torque = torquediff
			M.torqueTransmission = torquediff * engine.gears[M.gear]
			axleAV = M.engineAV / engine.gears[M.gear]
		end
	end
end

local function updateWheels(dt)
    local abs = math.abs
    local max = math.max
    local min = math.min

    local absABSavgAV = abs(ABSavgAV)
    --local absABSavgAV = abs(sensors.virtualAirspeed / wheelradius)
    local ABSwheelCount = 0

    M.avgAV = 0
    ABSavgAV = 0

    for _,wd in pairs(M.wheels) do
        local w = wd.obj
        wd.angularVelocity = w.angularVelocity
        wd.desiredBrakingTorque = 0
        local wheelAV = wd.angularVelocity * wd.wheelDir

        M.avgAV = M.avgAV + wheelAV

        -- braking
        if M.brake > 0 then
            wd.desiredBrakingTorque = wd.brakeTorque *
            (min(M.brake, wd.brakeInputSplit) + max(M.brake - wd.brakeInputSplit, 0) * wd.brakeSplitCoef)
            --anti-lock braking system (ABS)
            if wd.enableABSactuator then
                if abs(wd.angularVelocity) < absABSavgAV - wd.ABSthreshold then
                    if wd.ABStorque == -1 then wd.ABStorque = wd.desiredBrakingTorque end
                    wd.ABStorque = math.max(wd.ABStorque - wd.ABSchangeRate, 0)
                    wd.desiredBrakingTorque = wd.ABStorque
                    M.absPulse = bit.bxor(M.absPulse, 1)
                else
                    if wd.enableABSsensor then
                        ABSavgAV = ABSavgAV + wheelAV
                        ABSwheelCount = ABSwheelCount + 1
                    end
                    if wd.ABStorque ~= -1 then
                        wd.ABStorque = math.min(wd.ABStorque + wd.ABSchangeRate, wd.desiredBrakingTorque)
                        wd.desiredBrakingTorque = wd.ABStorque
                    end
                end
            else
                if wd.enableABSsensor then
                    ABSavgAV = ABSavgAV + wheelAV
                    ABSwheelCount = ABSwheelCount + 1
                end
            end
            wd.lastTorqueMode = 1
        else
            ABSavgAV = ABSavgAV + wheelAV
            ABSwheelCount = ABSwheelCount + 1
            wd.ABStorque = -1
            wd.lastTorqueMode = 0
        end

        -- and the parking brake
        if input.parkingbrake ~= 0 then
            wd.lastTorqueMode = 2
            wd.desiredBrakingTorque = max(wd.desiredBrakingTorque, wd.parkingTorque * input.parkingbrake)
        end
    end

    --update the ESC, right now it might override certain brake actions (handbrake for example), if a car has no esc, this is a nop call
    M.esc.update(dt)

    for _,wd in pairs(M.wheels) do
        wd.obj.brakingTorque = wd.brakePressureDelay:get(wd.desiredBrakingTorque) * wd.brakeThermalEfficiency
        wd.torque = 0 -- prepare torque for propulsion
    end

    M.avgAV = M.avgAV * invWheelCount
    ABSavgAV = ABSavgAV / ABSwheelCount

    avgDiffAV = 0
    for _,diff in pairs(diffs) do
        local w1 = diff.w1
        local w2 = diff.w2
        avgDiffAV = avgDiffAV + math.max(w1.angularVelocity * w1.wheelDir, w2.angularVelocity * w2.wheelDir) -- Use max wheel speed as the axle's speed
    end
    avgDiffAV = avgDiffAV * invDiffCount

    if v.data.engine and v.data.engine.gears then
        updateEngineState(dt)
    end

    -- propulsion
    local torque
    local absTorqueTransmission = abs(M.torqueTransmission)
    local absAxleAV = abs(axleAV)

    for _,diff in pairs(diffs) do
        local w1 = diff.w1
        local w2 = diff.w2
        local w1U = w1
        local w2U = w2
        local w1AV = w1.angularVelocity * w1.wheelDir
        local w2AV = w2.angularVelocity * w2.wheelDir
        local w1AVU = w1AV
        local w2AVU = w2AV

        local absW1AV = abs(w1AV)
        local absW2AV = abs(w2AV)
        local w1torque = 0
        local w2torque = 0
        local w1torqueU = 0
        local w2torqueU = 0

        if absW2AV < absW1AV then
            w1AV, w2AV = w2AV, w1AV
            absW1AV, absW2AV = absW2AV, absW1AV
            w1, w2 = w2, w1
        end

        -- closed diff
        if diff.state == 2 then
            if absW2AV > absAxleAV then
                torque = min((absW2AV - absAxleAV), 1) * M.torqueTransmission * diff.engineTorqueCoef
            else
                torque = M.torqueTransmission * diff.engineTorqueCoef
            end

            local ltorque
            diff.angle = diff.angle + (w1AVU - w2AVU) * dt
            if abs(diff.angle) < 1 then
                local absangle = abs(diff.angle)
                ltorque = fsign(diff.angle) * absangle * absangle * diff.closedTorque
            else
                local signangle = fsign(diff.angle)
                ltorque = signangle * diff.closedTorque
                diff.angle = signangle
            end

            local absltorque = math.abs(ltorque)
            local absTorque = math.abs(torque)
            if absltorque < absTorque then
                if fsign(ltorque) ~= fsign(torque) then
                    w1torqueU = torque
                    w2torqueU = torque + fsign(ltorque) * absltorque * 2
                else
                    w1torqueU = torque - fsign(ltorque) * absltorque * 2
                    w2torqueU = torque
                end
            else
                w1torqueU = -ltorque
                w2torqueU = ltorque
            end
        else
            -- Open diffs
            if absW2AV > absAxleAV - 0.1 then
                if absW2AV > absAxleAV then
                    local coef = min((absW2AV - absAxleAV) * 10, 1)
                    torque = coef * absTorqueTransmission * diff.engineTorqueCoef
                    w1torque = fsign(M.torqueTransmission) * torque
                    w2torque = fsign(axleAV - w2AV) * torque
                else
                    torque = (absW2AV - absAxleAV) * M.torqueTransmission * diff.engineTorqueCoef * 10
                    w1torque = torque
                    w2torque = torque
                end
            else
                torque = M.torqueTransmission * diff.engineTorqueCoef
                w1torque = torque
                w2torque = torque
            end
            w1.torque = w1.torque + w1torque * w1.wheelDir
            w2.torque = w2.torque + w2torque * w2.wheelDir

            -- LSD diffs
            if diff.state ~= 0 then
                local ltorque
                diff.angle = diff.angle + (w1AVU - w2AVU) * dt
                if abs(diff.angle) < 1 then
                    local absangle = abs(diff.angle)
                    ltorque = fsign(diff.angle) * absangle * absangle * diff.closedTorque
                else
                    local signangle = fsign(diff.angle)
                    ltorque = signangle * diff.closedTorque
                    diff.angle = 0
                end
                w1torqueU = -ltorque
                w2torqueU = ltorque
            else
                diff.angle = 0
            end
        end
        w1U.torque = w1U.torque + w1torqueU * w1U.wheelDir
        - fsign(w1U.angularVelocity) * min(abs(w1U.angularVelocity), 1) * axleFriction
        w2U.torque = w2U.torque + w2torqueU * w2U.wheelDir
        - fsign(w2U.angularVelocity) * min(abs(w2U.angularVelocity), 1) * axleFriction

        w1U.obj.torque = w1U.torque
        w2U.obj.torque = w2U.torque
    end

    M.axleAngle = M.axleAngle + avgDiffAV * dt
    M.driveshaftAngle = M.driveshaftAngle + avgDiffAV * dt * engineDifferential
end

local function disableEngine()
    if M.engineDisabled then return end
    M.engineDisabled = true
    beamstate.addDamage(100000)
    engIdleAV = 0
    gui.message("Engine damaged", 10, "damage.engine")
    warningState.engine = true
    sendWarningStateToUi()
end


local function checkEngine(dt)
    if M.engineDisabled or floodLevel > hydrolockThreshold then return end

    -- engine starts glooding if any of the waterDamage nodes is underwater
    local isFlooding = false
    for _, n in ipairs(v.data.engine.waterDamage.nodes) do
        if obj:inWater(n) then
            isFlooding = true
            break
        end
    end

    -- calculate flooding speed (positive) or drying speed (negative, and arbitrarily slower than flooding after some testing)
    local wetspeed = 1
    local dryspeed = -wetspeed/2
    local floodSpeed = (isFlooding and wetspeed or dryspeed) * (M.rpm/v.data.engine.maxRPM) -- TODO use torque instead of RPM (when torque calculation becomes more realistic)

    -- actual check for engine dying. in the future we may want to implement stalling too
    floodLevel = math.min(1,math.max(0,floodLevel + dt*floodSpeed))
    if floodLevel > hydrolockThreshold then
        gui.message("Engine hydrolocked", 10, "damage.flood")
        warningState.flood = true
        sendWarningStateToUi()
        scaleEngineFriction(10000) -- avoid piston movement, simulate broken connecting rods
        return
    end

    -- we compute the flooding percentage in steps of 10%...
    local currPercent = math.floor(0.5+floodLevel*10)*10
    -- ...and use that to check when to perform UI updates
    if currPercent ~= prevFloodPercent then
        if currPercent > prevFloodPercent then
            gui.message("Engine flooding with water ("..currPercent.."%)", 4, "damage.flood")
            warningState.floodPerc = currPercent
            sendWarningStateToUi()
        else
            if currPercent < 10 then
                gui.message("Engine dried up", 4, "damage.flood")
                warningState.flood = false
                sendWarningStateToUi()
            else
                gui.message("Engine drying from water ("..currPercent.."%)", 4, "damage.flood")
            end
        end
    end
    prevFloodPercent = currPercent
    --log("I", "", "Intake -> level: "..floodSpeed.." -> "..floodLevel..", level: "..currPercent)
end

local function simpleAutoGearbox(dt)
    local engine = v.data.engine
    local tmpCurGear = M.gear
    gearChangedTimer = math.max(gearChangedTimer - dt, 0 )

    if clutchRatio ~= 1 then
        gearChangedTimer = gearChangeDelay
    end
    -- automatic
    --interpolate based on throttle between high/low ranges
    local throttleCubed = M.throttle * M.throttle * M.throttle
    local shiftDown = engine.lowShiftDownAV + (engine.highShiftDownAV - engine.lowShiftDownAV) * throttleCubed
    local shiftUp   = engine.lowShiftUpAV + (engine.highShiftUpAV - engine.lowShiftUpAV) * throttleCubed

    if gearChangedTimer == 0 and input.clutch == 0 then
        local tmpEngineAV = M.engineAV
        local relEngineAV = M.engineAV / engine.gears[M.gear]
        while tmpEngineAV < shiftDown do
            M.gear = M.gear - sign(M.gear)
            gearChangedTimer = gearChangeDelay
            if M.gear == 0 or M.gear <= -engine.revGearCount or M.gear >= engine.fwdGearCount then break end
            tmpEngineAV = relEngineAV * engine.gears[M.gear]
        end
        while tmpEngineAV > shiftUp do
            M.gear = M.gear + sign(M.gear)
            gearChangedTimer = gearChangeDelay
            if M.gear == 0 or M.gear <= -engine.revGearCount or M.gear >= engine.fwdGearCount then break end
            tmpEngineAV = relEngineAV * engine.gears[M.gear]
        end
    end

    -- neutral gear handling
    if M.gear == 0 then
        if input.throttle > 0 and smoothedAvgAV > -1 and tmpCurGear >=0 then
            M.gear = 1
        end

        if input.brake > 0 and smoothedAvgAV <= 0.15 and tmpCurGear <=0 then
            M.gear = -1
        end
    end

    -- Control clutch to buildup engine RPM
    if input.clutch == 0 then
        clutchRatio = math.min(math.max((M.engineAV - engIdleAV)/(shiftDown - engIdleAV),0),1)
    end

    M.gear = math.min(math.max(M.gear, -engine.revGearCount), engine.fwdGearCount)
end

local function realAutoGearbox(dt)
    if clutchRatio ~= 1 then
        gearChangedTimer = gearChangeDelay
    end

    clutchRatio = 1
    gearChangedTimer = math.max( gearChangedTimer - dt, 0 )

    local engine = v.data.engine

    M.shifterPosition = math.max( -2, math.min(3, M.shifterPosition))

    if M.shifterPosition == 0 then
        M.gear = 0
        return
    end

    if math.abs(M.shifterPosition) == 1 then
        --interpolate based on throttle between high/low ranges
        local throttleCubed = M.throttle * M.throttle * M.throttle
        local shiftDown = engine.lowShiftDownAV + (engine.highShiftDownAV - engine.lowShiftDownAV) * throttleCubed
        local shiftUp   = engine.lowShiftUpAV + (engine.highShiftUpAV - engine.lowShiftUpAV) * throttleCubed
        if gearChangedTimer == 0 then
            local tmpEngineAV = M.engineAV
            local relEngineAV = M.engineAV / engine.gears[M.gear]

            while tmpEngineAV < shiftDown do
                M.gear = M.gear - sign(M.gear)
                gearChangedTimer = gearChangeDelay
                if M.gear == 0 or M.gear <= -engine.revGearCount or M.gear >= engine.fwdGearCount then break end
                tmpEngineAV = relEngineAV * engine.gears[M.gear]
            end
            while tmpEngineAV > shiftUp do
                M.gear = M.gear + sign(M.gear)
                gearChangedTimer = gearChangeDelay
                if M.gear == 0 or M.gear <= -engine.revGearCount or M.gear >= engine.fwdGearCount then break end
                tmpEngineAV = relEngineAV * engine.gears[M.gear]

            end

            if M.shifterPosition == 1 then M.gear = math.max(1, M.gear) end
            if M.shifterPosition == -1 then M.gear = math.min(-1, M.gear) end
            -- Control clutch to buildup engine RPM

        end
        if math.abs(M.gear) == 1 and clutchRatio == 1 then
            clutchRatio = math.min(math.max(M.engineAV/shiftDown,0),1)
        end
    elseif M.shifterPosition == 3 then
        M.gear = 1
    elseif M.shifterPosition == 2 then
        M.gear = 2
    elseif M.shifterPosition == -2 then
        M.gear = 0
        M.brake = 1
    end

    M.gear = math.min(math.max(M.gear, -engine.revGearCount), engine.fwdGearCount)
end

local function updateEngine(dt)
    local engine = v.data.engine
    engineState = 1

    --Update GUI engine RPM
    M.rpm = M.engineAV * 9.5493

    M.axleAngle = M.axleAngle % math.pi
    M.driveshaftAngle = M.driveshaftAngle % math.pi

    M.throttle = math.min(math.max(input.throttle, 0), 1)
    M.brake = math.min(math.max(input.brake, 0), 1)
    clutchRatio = 1 - math.min(math.max(input.clutch, 0), 1)

    local maxCurrentTorque = (engine.torqueCurve[math.floor(M.rpm)] or 0) - M.engFriction - (M.brakingCoef * M.rpm * 0.1047197177)
    M.engineLoad = math.max(loadSmoother:get(M.torque / ((maxCurrentTorque + 1e-30) * M.turbocharger.turboCoefPower), dt), 0)

    --Activate the RPM limiter if we are initially above the maxAV of the engine
    if revLimiterActive or M.engineAV > v.data.engine.maxAV then
        engineState = engineState * 0
        --Deactivate the limiter once below the deactivation threshold
        revLimiterActive = not (M.engineAV <= revLimiterAVThreshold)
    end

    M.engine.updateGFX(dt)
    M.wheelSimulation.updateGFX(dt)

    -- Fuel consumption
    M.fuel = M.fuel - M.engineWorkPerUpdate * invBurnEfficiencyFuelDensity - M.fuelLeakRate * dt
    for k,v in pairs(fuelNodes) do
        obj:setNodeMass(k, v + M.fuel * fuelNodeWeightCoef)
    end
    M.engineWorkPerUpdate = 0
    M.frictionLossPerUpdate = 0
    M.pumpingLossPerUpdate = 0

    smoothedAvgAV = avgAVsmoother:get(M.avgAV)

    if M.fuel <= 0 then
        M.fuel = 0
        engineState = 0
    end

    if M.manualShift then
        if M.autoClutch then
            if M.brake ~= 0 and M.engineAV < engIdleAV  * 2 then
                clutchRatio = 0
            end

            if M.engineAV < engStallAV then
                clutchRatio = 0
            end
        end
    else
        -- driving backwards? - only with automatic shift - for obvious reasons ;)
        if M.shifterMode == 2 and M.gear <= 0 and smoothedAvgAV <= 0.1 and freeze == 0 then
            M.throttle, M.brake = M.brake, M.throttle
        end
    end

    -- check for engine disabling (every 0.1 sec, takes around 0.000002 sec to run)
    checkEngineTimer = checkEngineTimer + dt
    local checkEnginePeriod = 0.1
    if checkEngineTimer > checkEnginePeriod then
        checkEngineTimer = checkEngineTimer - checkEnginePeriod
        checkEngine(checkEnginePeriod)
    end

    if M.engineDisabled then
        engineState = 0
        M.rpm = 0
    end

    local prevGear = M.gear
    --transmission logic
    if M.manualShift then
        -- manual
        M.shifterPosition = math.min( math.max(M.shifterPosition, -engine.revGearCount), engine.fwdGearCount )
        M.gear = M.shifterPosition
        if prevGear ~= M.gear and M.autoClutch then
            clutchTimer = clutchDuration
        end
    else
        if M.shifterMode == 2 then simpleAutoGearbox(dt)
        elseif M.shifterMode == 3 then realAutoGearbox(dt)
		elseif M.shifterMode == 4 then realAutoGearbox(dt)
        end
    end

    if freeze == 1 then
        M.brake = 1
        clutchRatio = 0
        clutchTimer = 0
        if not (M.manualShift and not M.autoClutch) then
            M.shifterPosition = 0
            M.gear = 0
        end
    end

    if ((clutchDuration ~= 0 and clutchTimer == clutchDuration) or (gearChangeDelay ~= 0 and gearChangedTimer == gearChangeDelay)) and prevGear < M.gear then
        M.throttle = M.throttle * 0.5
    end

    M.turbocharger.updateGFX(dt)
    M.esc.updateGFX(dt)

    M.clutch = 1 - clutchRatio
    M.engineThrottle = M.throttle * engineState * M.throttleFactor
end

local function disableDriveshaft()
    axleFriction = 0
    viscousCoupling = 0
    if M.driveshaftDisabled then
        return
    end
    M.driveshaftDisabled = true
    beamstate.addDamage(100000)
    gui.message("Driveshaft broken", 10, "damage.driveshaft")
    warningState.driveshaft = true
    sendWarningStateToUi()
end


local function updateCounts()
    propWheelsCount = 0
    M.wheelCount = 0
    for _,wd in pairs(M.wheels) do
        if wd.propulsed ~= nil and wd.propulsed > 0 then
            propWheelsCount = propWheelsCount + 1
        end
        M.wheelCount = M.wheelCount + 1
    end

    if M.wheelCount ~= 0 then
        invWheelCount = 1 / M.wheelCount
    else
        invWheelCount = 0
    end
end

local function updateDifferentials()
    local wheelNames = {}
    local wheels = {}

    -- Clear old diffs config state
    for _,diff in pairs(diffs) do
        diff.w1.obj.torque = 0
        diff.w2.obj.torque = 0
    end

    for _,wd in pairs(M.wheelInfo) do
        wheelNames[wd.name] = M.wheels[wd.wheelID]
        wheels[wd.name] = wd
    end


    local totalEngCoef = 0

    -- Sum engine torque coefs
    v.data.differentials = v.data.differentials or {}

    for _, diff in pairs(v.data.differentials) do
        if wheelNames[diff.wheelName1] and wheelNames[diff.wheelName2] then
            diff.engineTorqueCoef = diff.engineTorqueCoef or 1
            totalEngCoef = totalEngCoef + diff.engineTorqueCoef
        end
    end

    -- Update diffs
    local i = 0
    diffs = {}
    for _,diff in pairs(v.data.differentials) do
        if wheelNames[diff.wheelName1] and wheelNames[diff.wheelName2] then
            diffs[i] = {    w1 = wheelNames[diff.wheelName1] ,
                w2 = wheelNames[diff.wheelName2] ,
                dtype = diffTypeMap[string.lower(diff.type or "open")] ,
                closedTorque = diff.closedTorque ,
                state = diffStateMap[string.lower(diff.state or "open")] ,
                engineTorqueCoef = 0.5 * diff.engineTorqueCoef / totalEngCoef, --per diff wheel (0.5)
                angle = 0
            }
            wheels[diff.wheelName1].hasDiffAttached = true
            wheels[diff.wheelName2].hasDiffAttached = true
            i = i + 1
        end
    end

    diffCount = i
    if diffCount ~= 0 then
        invDiffCount = 1 / diffCount
    else
        invDiffCount = 0
    end
end

local function beamBroke(id)
    M.engine.beamBroke(id)

    if not v.data.beams[id].name then return end
    local doUpdate = false
    local brokenBeamName = v.data.beams[id].name

    if v.data.engine then
        if v.data.engine.onBeamBreakDisableEngine and engineBeamBreakLookup[brokenBeamName] then
            disableEngine()
        end

        if v.data.engine.onBeamBreakDisableDriveshaft and driveshaftBeamBreakLookup[brokenBeamName] then
            disableDriveshaft()
        end

        if v.data.engine.fuelTankBeams and fuelTankBeamBreakLookup[brokenBeamName] then
            gui.message("Fuel leaking", 10, "damage.fuelleak")
            warningState.fuelleak = true
            sendWarningStateToUi()
            M.fuelLeakRate = M.fuelLeakRate + 1 / #v.data.engine.fuelTankBeams
        end
    end

    for n,wd in pairs(M.wheels) do
        if wd.axleBeams and wd.axleBeamLookup[brokenBeamName] and not driveshaftBeamBreakLookup[brokenBeamName] then
            gui.message(wd.name.." Wheel axle broken", 10, "damage.wheelaxle"..wd.name)
            warningState["wheelaxle"..wd.name] = true
            sendWarningStateToUi()
            doUpdate = true
            local w = wd.obj
            w.torque = 0
            w.brakingTorque = 0
            M.wheels[n] = nil
        end
    end

    if doUpdate then
        updateCounts()
        updateDifferentials()
    end
end

local function updateWheelSlip(p)
    if not p then return end
    if not v.data.nodes[p.id1] then return end
    local wheelID = v.data.nodes[p.id1].wheelID
    if wheelID then
        local wInfo = M.wheelInfo[wheelID]
        wInfo.lastSlip = math.max(p.slipVel, wInfo.lastSlip)
        -- Smoothed instant energy (E/dt)
        wInfo.slipEnergy = 0.5 * (p.slipForce * p.slipVel + wInfo.slipEnergy)
        wInfo.contactMaterialID1 = p.materialID1
        wInfo.contactMaterialID2 = p.materialID2
        wInfo.contactDepth = math.max (p.depth, wInfo.contactDepth)
        wInfo.downForceRaw = wInfo.downForceRaw - p.normalForce
    end
end

local function updateWheelsGFX(dt)
    for _,wd in pairs(M.wheelInfo) do
        wd.contactMaterialID1 = -1
        wd.contactMaterialID2 = -1
        wd.lastSlip = 0
        wd.slipEnergy = 0
        wd.contactDepth = 0
        wd.downForce = wd.downForceSmoother:get(wd.downForceRaw)
        wd.downForceRaw = 0

        wd.brakeEnergyPerUpdate = wd.brakeEnergyPerUpdate + math.abs(wd.obj.brakingTorque * math.abs(wd.obj.angularVelocityBrakeCouple) * dt)
    end

    M.absPulse = 0
end

local function getShifterModeName(mode)
    if mode == 0 then return "Manual" end
    if mode == 1 then return "Manual (clutch assist)" end
    if mode == 2 then return "Arcade" end
    if mode == 3 then return "Automatic" end
	if mode == 4 then return "Realistic Automatic" end
    return "Unknown shifting mode"
end
local function getShifterModeIndex(mode)
    for i,m in ipairs(activeModes) do
        if m == mode then return i end
    end
    return nil
end

local function setShifterMode(v)
    v = v or 0

    M.shifterPosition = M.gear

    M.activeModeIndex = v

    -- limits
    if M.activeModeIndex > #activeModes then
        M.activeModeIndex = 1
    end
    if M.activeModeIndex < 1 then
        M.activeModeIndex = 1
    end

    M.shifterMode = activeModes[M.activeModeIndex]

    -- now do the important things
    if M.announcedShifterMode ~= M.shifterMode then
        gui.message("Using '"..getShifterModeName(M.shifterMode).."' shifting mode", 5, "shiftermode")
        M.announcedShifterMode = M.shifterMode
    end
    if M.shifterMode == 0 then
        -- full manual
        M.autoClutch = false
        M.manualShift = true
    elseif M.shifterMode == 1 then
        -- manual with automatic clutch
        M.autoClutch = true
        M.manualShift = true
    elseif M.shifterMode == 2 then
        -- simple automatic
        M.autoClutch = true
        M.manualShift = false
        viscousCoupling = M.driveshaftDisabled and 0 or viscousCouplingOrig
    elseif M.shifterMode == 3 then
        -- real automatic
        M.autoClutch = true
        M.manualShift = false
        viscousCoupling = M.driveshaftDisabled and 0 or viscousCouplingOrig
        M.shifterPosition = 0
	elseif M.shifterMode == 4 then
        -- more realistic automatic with tc and lockup (auntie 2016)
        M.autoClutch = true
        M.manualShift = false
        viscousCoupling = M.driveshaftDisabled and 0 or viscousCouplingOrig
        M.shifterPosition = 0
    end

    -- if not automatic
    if M.shifterMode < 2 and not M.driveshaftDisabled then
        viscousCoupling = math.min(clutchTorque / maxGearRatio, (engInertia / obj:getPhysicsDt()))
    end
end

local function toggleShifterMode()
    setShifterMode(M.activeModeIndex + 1)
end

local function smartShifterModeHShifter(gear)
    local newModeIndex = nil
    if gear == 0 then return end
    if gear > 3 then
        if not newModeIndex then newModeIndex = getShifterModeIndex(0) end -- manual full
        if not newModeIndex then newModeIndex = getShifterModeIndex(1) end -- manual autoclutch
        if not newModeIndex then gui.message("Cannot shift to gear #"..gear.." with an automatic gearbox", 5, "shiftermode") end
    else
        if not newModeIndex then newModeIndex = getShifterModeIndex(3) end -- real auto
        if not newModeIndex then newModeIndex = getShifterModeIndex(0) end -- manual full
        if not newModeIndex then newModeIndex = getShifterModeIndex(1) end -- manual autoclutch
    end
    if not newModeIndex then return end
    setShifterMode(newModeIndex)
    gui.message("H-shifter has been used. Now using: '"..getShifterModeName(activeModes[newModeIndex]).."'", 10, "shiftermode")
end
local function smartShifterModePaddles()
    local newModeIndex = nil
    if not newModeIndex then newModeIndex = getShifterModeIndex(3) end -- real auto
    if not newModeIndex then newModeIndex = getShifterModeIndex(1) end -- manual autoclutch
    if not newModeIndex then newModeIndex = getShifterModeIndex(0) end -- manual full
    if not newModeIndex then return end
    setShifterMode(newModeIndex)
    gui.message("Shifter has been used. Now using: '"..getShifterModeName(activeModes[newModeIndex]).."'", 10, "shiftermode")
end

local function shiftUp()
    if M.shifterMode == 2 then smartShifterModePaddles() end -- from arcade auto to a suitable mode
    M.shifterPosition = M.shifterPosition + 1
end

local function shiftDown()
    if M.shifterMode == 2 then smartShifterModePaddles() end -- from arcade auto to a suitable mode
    M.shifterPosition = M.shifterPosition - 1
end
local function shiftToGear(val)
    -- when in full automatic, shifting only to -1, 1, 2 and N is allowed
    -- TODO: lock -1 when moving forwards
    if M.shifterMode == 2 or (M.shifterMode == 3 and val > 3) then smartShifterModeHShifter(val) end
    M.shifterPosition = val
end

local function setDifferentialMode(v)
    M.differentialMode = math.min(math.max(v or 0, 0), 2)

    for _,diff in pairs(diffs) do
        if diff.dtype == diffTypeMap.lsd then
            diff.state = M.differentialMode
        end
    end
end

local function toggleDifferentialMode()
    setDifferentialMode((M.differentialMode + 1) % 3)
end

local function genDiffs()
    diffs = {}

    if propWheelsCount ~= 2 then
        return
    end

    local diffWheels = {}
    local wn = 0

    for _,wd in pairs(M.wheels) do
        if wd.propulsed ~= nil and wd.propulsed > 0 then
            diffWheels[wn] = wd
            wn = wn + 1
        end
    end

    if wn == 2 then
        log('D', "drivetrain.genDiffs", "Found 2 propulsed wheels: generating Open Differential between them")
        diffs[0] = {w1 = diffWheels[0], w2 = diffWheels[1], dtype = diffTypeMap['open'],
            state = 0, engineTorqueCoef = 0.5, angle = 0}
    end
end

local function toggleESCMode()
    M.esc.toggleESCMode()
end

M.updateEngine = nop
M.updateWheels = nop
M.updateWheelSlip = nop
M.resetWheels = nop
M.beamBroke = nop

local function init()
    M.engineAV = 0
    axleAV = 0
    clutchTimer = 0
    gearChangedTimer = 0
    M.engineDisabled = false
    M.engineIgnition = true
    M.engineWorkPerUpdate = 0
    M.frictionLossPerUpdate = 0
    M.pumpingLossPerUpdate = 0
    M.gear = 0
    M.shifterPosition = 0
    M.fuelLeakRate = 0
    M.torqueTransmission = 0
    avgPropAV = 0
    M.avgAV = 0
    M.axleAngle = 0
    M.driveshaftAngle = 0
    M.throttleFactor = 1
    M.engineLoad = 0
    M.driveshaftDisabled = false
    engineState = 1

    M.wheels = {}
    M.wheelInfo = {}

    M.updateWheels = updateWheels

    warningState = {}
    sendWarningStateToUi()

    avgAVsmoother:reset()
    floodLevel = 0
    prevFloodPercent = 0

    if not v or not v.data then
        log('E', 'drivetrain.init', 'Init aborted due to errors')
        return
    end

    -- init the functions correctly
    M.updateEngine = updateEngine
    M.updateWheels = updateWheels
    M.updateWheelSlip = updateWheelSlip
    M.updateWheelsGFX = updateWheelsGFX
    M.beamBroke = beamBroke

    if v.data.wheels and tableSize(v.data.wheels) > 0 then
        M.wheels = shallowcopy(v.data.wheels)
        M.wheelInfo = shallowcopy(v.data.wheels)
    else
        M.updateWheels = nop
        M.updateWheelSlip = nop
    end

    local pressureSmoothWindow = 0
    local stepsForABSrate
    if v.data.engine ~= nil then
        stepsForABSrate = 2000 / (v.data.engine.ABSrate or 15)
        pressureSmoothWindow = (v.data.engine.brakePressureDelay or 0.1) * 2000
    end

    local wobj
    for _,wd in pairs_safe(M.wheelInfo) do
        wobj = obj:getWheel(wd.wheelID)
        if wobj then
            wd.lastTorqueMode = 0
            M.wheelInfo[wd.wheelID].lastSlip = 0
            M.wheelInfo[wd.wheelID].slipEnergy = 0
            M.wheelInfo[wd.wheelID].contactMaterialID1 = -1
            M.wheelInfo[wd.wheelID].contactMaterialID2 = -1
            M.wheelInfo[wd.wheelID].contactDepth = 0
            M.wheelInfo[wd.wheelID].downForceRaw = 0
            M.wheelInfo[wd.wheelID].downForceSmoother = newExponentialSmoothing(30)
            M.wheelInfo[wd.wheelID].slipEnergySmoother = newTemporalSmoothing()
            M.wheelInfo[wd.wheelID].downForce = 0
            M.wheelInfo[wd.wheelID].obj = wobj
            M.wheels[wd.wheelID].obj = wobj
            if M.wheels[wd.wheelID].enableABS ~= nil and M.wheels[wd.wheelID].enableABS then
                M.wheels[wd.wheelID].enableABSsensor = true
            else
                M.wheels[wd.wheelID].enableABSsensor = M.wheels[wd.wheelID].enableABSsensor or false
            end
            M.wheels[wd.wheelID].enableABSactuator = M.wheels[wd.wheelID].enableABSactuator or M.wheels[wd.wheelID].enableABS or false
            M.wheels[wd.wheelID].ABSthreshold = M.wheels[wd.wheelID].ABSthreshold or 5
            M.wheels[wd.wheelID].brakeTorque = M.wheels[wd.wheelID].brakeTorque or 0
            M.wheels[wd.wheelID].ABStorque = 0
            M.wheels[wd.wheelID].ABSchangeRate = M.wheels[wd.wheelID].brakeTorque / stepsForABSrate
            M.wheels[wd.wheelID].parkingTorque = M.wheels[wd.wheelID].parkingTorque or 0
            M.wheels[wd.wheelID].torque = 0
            M.wheels[wd.wheelID].angularVelocity = 0
            M.wheels[wd.wheelID].brakeInputSplit = math.max(math.min(M.wheels[wd.wheelID].brakeInputSplit or 1, 1), 0)
            M.wheels[wd.wheelID].brakeSplitCoef = math.max(math.min(M.wheels[wd.wheelID].brakeSplitCoef or 1, 1), 0)
            M.wheels[wd.wheelID].brakePressureDelay = newExponentialSmoothing(pressureSmoothWindow)
            M.wheels[wd.wheelID].brakeEnergyPerUpdate = 0
            M.wheels[wd.wheelID].brakeThermalEfficiency = 1
            wheelradius = wd.radius
            M.wheels[wd.wheelID].axleBeamLookup = {}
            if M.wheels[wd.wheelID].axleBeams then
                for _,name in pairs(M.wheels[wd.wheelID].axleBeams) do
                    M.wheels[wd.wheelID].axleBeamLookup[name] = true
                end
            end
        else
            log('W', "drivetrain.init", 'Wheel "'..M.wheels[wd.wheelID].name..'" could not be added to drivetrain')
            for _,v in pairs(v.data.nodes) do
                if v.wheelID == wd.wheelID then
                    v.wheelID = nil
                end
            end
            M.wheels[wd.wheelID] = nil
            M.wheelInfo[wd.wheelID] = nil
        end
    end

    updateCounts()

    if v.data.differentials == nil then
        genDiffs()
    else
        updateDifferentials()
    end

    if v.data.engine and v.data.engine.gears then
        clutchDuration = v.data.engine.clutchDuration or 0.5
        invClutchDuration = 1 / clutchDuration
        engIdleAV = v.data.engine.idleAV
        engStallAV = v.data.engine.stallAV or engIdleAV * 0.5
        torsionCoef = v.data.engine.torsionCoef or 40
        engInertia = v.data.engine.inertia or 0.1
        engineDifferential = v.data.engine.differential
        local revLimiterAVDrop = (v.data.engine.revLimiterRPMDrop or (v.data.engine.maxRPM * 0.03)) * 0.1047197551
        revLimiterAVThreshold = math.min(v.data.engine.maxAV - revLimiterAVDrop, v.data.engine.maxAV)

        if v.data.engine.torqueReactionNodes_nodes then
            local t = v.data.engine.torqueReactionNodes_nodes
            if #t == 3 and v.data.nodes[t[1]] ~= nil and v.data.nodes[t[2]] ~= nil and v.data.nodes[t[3]] ~= nil then
                torqueReactionNodes = {t[1], t[2], t[3]}
            else
                torqueReactionNodes = nil
            end
        end

        if v.data.engine.transmissionType == "automatic" then
            activeModes = autoModes
        end

        if v.data.engine.transmissionType == "manual" then
            activeModes = manualModes
        end

        if engInertia ~= 0 then
            invEngInertia = 1 / engInertia
        else
            invEngInertia = 10
        end
        halfInvEngInertia = invEngInertia * 0.5
        -- fuel
        M.fuelCapacity = v.data.engine.fuelCapacity or 100
        M.fuel = M.fuelCapacity
        local fuelEnergyDensity = (v.data.engine.fuelEnergyDensity or 32) * 1000000 -- MJ/liter
        burnEfficiency = v.data.engine.burnEfficiency or 1
        if fuelEnergyDensity * burnEfficiency ~= 0 then
            invBurnEfficiencyFuelDensity = 1 / (fuelEnergyDensity * burnEfficiency)
        else
            invBurnEfficiencyFuelDensity = 1
        end
        local fuelNodeCount = 0
        fuelNodes = {}
        fuelNodeWeightCoef = 0
        if v.data.engine.fuel and v.data.engine.fuel._engineGroup_nodes then
            for _,n in pairs(v.data.engine.fuel._engineGroup_nodes) do
                fuelNodes[n] = v.data.nodes[n].nodeWeight --save initial mass as the offset for the fuel node weights
                fuelNodeCount = fuelNodeCount + 1
            end
            if fuelNodeCount > 0 then
                fuelNodeWeightCoef = 0.75 / fuelNodeCount --generic density of 0.75kg/m^3
            end
        end

        driveshaftBeamBreakLookup = {}
        if v.data.engine.onBeamBreakDisableDriveshaft then
            for _,name in pairs(v.data.engine.onBeamBreakDisableDriveshaft) do
                driveshaftBeamBreakLookup[name] = true
            end
        end

        engineBeamBreakLookup = {}
        if v.data.engine.onBeamBreakDisableEngine then
            for _,name in pairs(v.data.engine.onBeamBreakDisableEngine) do
                engineBeamBreakLookup[name] = true
            end
        end

        fuelTankBeamBreakLookup = {}
        if v.data.engine.fuelTankBeams then
            for _,name in pairs(v.data.engine.fuelTankBeams) do
                fuelTankBeamBreakLookup[name] = true
            end
        end

        diffRatio = v.data.engine.diffRatio or 2
        invDiffRatio = 1 / diffRatio
        M.engFriction = v.data.engine.friction or v.data.engine.engineFriction or 20
        M.brakingCoef = v.data.engine.brakingCoef or 0
        if v.data.engine.brakingCoefRPS ~= nil then
            M.brakingCoef = v.data.engine.brakingCoefRPS / ( 2 * math.pi )
        end
        axleFriction = v.data.engine.axleFriction or v.data.engine.axelFriction or 0

        if v.data.engine.torqueCurve == nil then
            v.data.engine.torqueCurve = {}
        end
        idleTorque = v.data.engine.torqueCurve[v.data.engine.idleRPM] or 0

        maxGearRatio = v.data.engine.maxGearRatio or 0
        gearChangeDelay = v.data.engine.autoGearChangeDelay or 0.75
        maxTorque = v.data.engine.maxTorque or 0
        maxPower = v.data.engine.maxPower or 0

        clutchTorque = v.data.engine.clutchTorque or v.data.engine.torqueLimit
        if not clutchTorque then
            clutchTorque = v.data.engine.maxTorque
            --log('D', "drivetrain.init", "clutchTorque not set. Set clutchTorque = "..clutchTorque)
        end

        viscousCoupling = v.data.engine.viscousCoupling or viscousCouplingOrig
        if not viscousCoupling then
            viscousCoupling = (clutchTorque / maxGearRatio) or 3
            --log('D', "drivetrain.init", "viscousCoupling not set. Set viscousCoupling = "..viscousCoupling)
        end
        viscousCoupling = math.min(viscousCoupling, (engInertia / obj:getPhysicsDt()))
        viscousCouplingOrig = viscousCoupling

        setShifterMode(M.activeModeIndex)
    else
        M.updateEngine = nop
        viscousCoupling = viscousCoupling or 3
        viscousCouplingOrig = viscousCoupling
    end

    M.engine = require("engine")
    M.engine.init()

    --M.wheelSimulation = require("wheelSimulation")
    --M.wheelSimulation.init()

    --Init turbo code
    if v.data.turbocharger then
        M.turbocharger = require("turbocharger")
        M.turbocharger.init()
    end

    --Init the esc module in case we have an escConfig in the car's jbeam
    if v.data.escConfig then
        M.esc = require("esc")
        M.esc.init()
    end

    -- apply settings if available
    --[[ THIS IS NOT WORKING YET
    if settings then
        local shifterMode = settings.getValue('gameplayDefaultShifterMode')
        print('#########################################################')
        print(type(shifterMode))
        print(tostring(shifterMode))
        if shifterMode ~= nil then
            setShifterMode(shifterMode)
        end
    end
    --]]
end

local function reset()
    -- for now: call init again, as it does no harm
    init()
end

local function refill()
    if M.fuel < M.fuelCapacity then
        M.fuel = math.min(M.fuel + 2, M.fuelCapacity)
    end
end

local function onDeserialized()
    setShifterMode(M.activeModeIndex)
end

local function setFreezeMode(m)
    freeze = m
end

-- public interface
M.reset           = reset
M.init            = init
M.refill          = refill

M.shiftUp         = shiftUp
M.shiftDown       = shiftDown
M.shiftToGear     = shiftToGear
M.toggleShifterMode = toggleShifterMode
M.toggleESCMode = toggleESCMode
M.setShifterMode  = setShifterMode
M.onDeserialized  = onDeserialized
M.toggleDifferentialMode = toggleDifferentialMode
M.setFreezeMode   = setFreezeMode
M.scaleEngineFriction = scaleEngineFriction
M.scaleEngineState = scaleEngineState

M.getWarningStateForUi = sendWarningStateToUi
return M
